package com.deryckwilliams.botControl1.android;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Set;
import java.util.UUID;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.SharedPreferences;
import android.os.AsyncTask;
import android.preference.ListPreference;
import android.preference.PreferenceManager;

import com.deryckwilliams.botControl1.BotCommsInterface;
import com.deryckwilliams.botControl1.BotController;
import com.deryckwilliams.botControl1.arduino.BotProtocol;
import com.deryckwilliams.botControl1.arduino.Cmd;
import com.deryckwilliams.botControl1.arduino.CmdVal;

/*******************************************************************************
 */
public final class MyBlueConnection extends AsyncTask<Object, Integer, Boolean> implements BotCommsInterface {
	private BotController myBotCtrl;
	private BotCtrlUI myUI;

	private BluetoothDevice myDevice = null;
	private BluetoothSocket mySock = null;
    private InputStream inStrm = null;
    private OutputStream outStrm = null;
    
    private boolean msgWaiting;
    private StringBuffer inBuffer;
    
	private final static UUID SPP_UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB");

	/***************************************************************************
	 */
	public MyBlueConnection(BotCtrlUI botUI) {
		super();
		this.myUI = botUI;
		myBotCtrl = new BotController(botUI, this);
	}
	
	/***************************************************************************
	 */
	public static void setupPrefs(ListPreference pref) {
		BluetoothAdapter mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
		
		if (mBluetoothAdapter == null) {
			pref.setEnabled(false);
			return;
		}

		if(!mBluetoothAdapter.isEnabled()) {
			pref.setEnabled(false);
			return;
		}
        
		Set<BluetoothDevice> pairedDevices = mBluetoothAdapter.getBondedDevices();
		// If there are paired devices
		if (pairedDevices.isEmpty()) {
			pref.setEnabled(false);
			return;
		}
		
	    // Loop through paired devices
		String entries[] = new String[pairedDevices.size()];
		int cnt = 0;
	    for (BluetoothDevice device : pairedDevices) {
	    	entries[cnt++] = device.getName();
	    }
		pref.setEnabled(true);
	    pref.setEntries(entries);
	    pref.setEntryValues(entries);
	    pref.setDefaultValue(entries[0]);
	}
	
	/***************************************************************************
	 */
	public boolean btInit() {
		BluetoothAdapter mBluetoothAdapter = BluetoothAdapter.getDefaultAdapter();
		
		if (mBluetoothAdapter == null) {
			myUI.setCommDisp(R.string.no_adapter);
			return false;
		}

		if(!mBluetoothAdapter.isEnabled()) {
			myUI.setCommDisp(R.string.bt_disabled);
			return false;
		}
        
		Set<BluetoothDevice> pairedDevices = mBluetoothAdapter.getBondedDevices();
		// If there are paired devices
		if (pairedDevices.size() == 0) {
			myUI.setCommDisp(R.string.no_bt_device);
			return false;
		}
		
		SharedPreferences prefs = PreferenceManager.getDefaultSharedPreferences(myUI);
        String devicePref = prefs.getString("btDevice", "");
        if(devicePref.length() == 0) {
			myUI.setCommDisp(R.string.bt_not_configured);
			return false;
        }
		
	    // Loop through paired devices
	    for (BluetoothDevice device : pairedDevices) {
	        // Add the name and address to an array adapter to show in a ListView
	    	if(device.getName().equals(devicePref)) {
	    		myDevice = device;
	    		break;
	    	}
	    }
	    
	    if(myDevice == null) {
	    	myUI.setCommDisp(R.string.not_paired);
			return false;
	    }
	    	    
	    return true;
	}

	/***************************************************************************
	 */
	public void disconnect() {
		myBotCtrl.disconnect();
	}

	/***************************************************************************
	 */
	public boolean isConnected() {
		return myBotCtrl.isConnected();
	}

	/***************************************************************************
	 */
	@Override
	protected Boolean doInBackground(Object... obj) {
		if(myDevice == null)
		{
			return false;
		}

        try {
        	mySock = myDevice.createRfcommSocketToServiceRecord(SPP_UUID);
        } catch (IOException e) {
        	myUI.setCommDisp(R.string.socket_fail);
			return false;
        }

        myUI.setCommDisp(R.string.connecting);
        try {
            // Connect the device through the socket. This will block
            // until it succeeds or throws an exception
        	mySock.connect();
        } catch (IOException connectException) {
            // Unable to connect; close the socket and get out
            try {
            	mySock.close();
            	mySock = null;
            } catch (IOException closeException) { }
            myUI.setCommDisp(R.string.contect_fail);
			return false;
        }

        myUI.setCommDisp(R.string.creating_streams);
        try {
            inStrm = mySock.getInputStream();
            outStrm = mySock.getOutputStream();
        } catch (IOException e) {
            try {
            	mySock.close();
            	mySock = null;
            } catch (IOException closeException) { }
        	myUI.setCommDisp(R.string.stream_create_fail);
			return false;
        }

        myUI.setCommDisp(R.string.connected);
        publishProgress(new Integer(1));
        msgWaiting = false;
        myBotCtrl.processingLoop();
        
        myUI.setCommDisp(R.string.disconnecting);
		if(outStrm != null)
		{
			try {
				outStrm.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			outStrm = null;
		}

		if(inStrm != null)
		{
			try {
				inStrm.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			inStrm = null;
		}

		if(mySock != null)
		{
			try {
				mySock.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			mySock = null;
		}

        myUI.setCommDisp(R.string.disconnected);
		return true;
	}

	/***************************************************************************
	 * @throws IOException 
	 */
	public synchronized void sendCmd(Cmd cmd) throws IOException
	{
		outStrm.write(BotProtocol.STX); 
		outStrm.write(cmd.chCmd);
		outStrm.write(cmd.chSubCmd);
		outStrm.write(BotProtocol.ETX);
	}
	
	/***************************************************************************
	 * @throws IOException 
	 */
	public synchronized void sendCmd(CmdVal cmdVal) throws IOException
	{
		String valStr = Integer.toString(cmdVal.iVal);
		outStrm.write(BotProtocol.STX);
		outStrm.write(cmdVal.chCmd);
		outStrm.write(cmdVal.chSubCmd);
		for(int cnt = valStr.length(); cnt < BotProtocol.CMD_VAL_LEN; cnt++)
			outStrm.write('0');
		outStrm.write(valStr.getBytes());
		outStrm.write(BotProtocol.ETX);
	}
	
	/***************************************************************************
	 */
	@Override
	protected void onPostExecute(Boolean result) {
		myUI.connectionDone();
	}

	/***************************************************************************
	 */
	@Override
	protected void onProgressUpdate(Integer... values) {
		myUI.connectionDone();
	}

	/***************************************************************************
	 */
	public void checkRecvMsg() {
		if(!msgWaiting) {
			try {
				checkMsg();
			} catch (IOException e) {}
		}
		if(msgWaiting) {
			processMsg();
		}
	}
	
	/***************************************************************************
	 */
	private void checkMsg() throws IOException {
		int val;

		while (!msgWaiting && inStrm.available() > 0) {
			val = inStrm.read();
			if(val == -1)
				return;
			switch((char)val) {
			case BotProtocol.STX:
				inBuffer = new StringBuffer();
				break;
			case BotProtocol.ETX:
				if (inBuffer != null && inBuffer.length() > 0)
					msgWaiting = true;
				break;
			case BotProtocol.ACK:
			case BotProtocol.NAK:
				break;
			default:
				inBuffer.append((char)val);
				break;
			}
		}
	}
	
	/*********************************************************************************
	 * Convert to an int the command received from the bot
	 */
	private int getCmdSensorVal() {
		try {
			return Integer.parseInt(inBuffer.substring(2, BotProtocol.CMD_VAL_LEN + 2), 10);
		} catch(Exception ex) {
			return 0;
		}
	}

	/*********************************************************************************
	 * Process the message as received from the bot
	 */
	private void processMsg() {
	  if(inBuffer.length() != (2 + BotProtocol.CMD_VAL_LEN))
		  return;

	  int val = getCmdSensorVal();
	  CmdVal cmd = new CmdVal(inBuffer.charAt(0), inBuffer.charAt(1), val);
	  myBotCtrl.recvMsg(cmd);
	  
	  // now the final common stuff
	  msgWaiting = false;
	}
}
